# Import built-in modules
from collections import namedtuple


VirusSignature = namedtuple("VirusSignature", ["name", "signature"])

# https://regex101.com/r/0MNzF7/1
virus20240430_sig1 = VirusSignature("virus20240430", "python(.*);.+exec.+(pyCode).+;")
# https://regex101.com/r/2D14UA/1
virus20240430_sig2 = VirusSignature("virus20240430", r"^\['.+']")



# Import built-in modules
from contextlib import contextmanager
import logging

# Import local modules
from maya_umbrella.cleaner import MayaVirusCleaner
from maya_umbrella.collector import MayaVirusCollector
from maya_umbrella.filesystem import get_hooks
from maya_umbrella.filesystem import load_hook
from maya_umbrella.i18n import Translator
from maya_umbrella.log import setup_logger
from maya_umbrella.maya_funs import is_maya_standalone
from maya_umbrella.maya_funs import om


# Global list to store IDs of Maya callbacks
MAYA_UMBRELLA_CALLBACK_IDS = []


def _add_callbacks_id(id_):
    """Add a callback ID to the global list if it's not already present.

    Args:
        id_ (int): ID of the callback to be added.
    """
    global MAYA_UMBRELLA_CALLBACK_IDS
    if id_ not in MAYA_UMBRELLA_CALLBACK_IDS:
        MAYA_UMBRELLA_CALLBACK_IDS.append(id_)


class MayaVirusDefender(object):
    """A class to defend against Maya viruses.

    Attributes:
        _vaccines (list): List to store vaccines.
        callback_maps (dict): Dictionary to map callback names to MSceneMessage constants.
        auto_fix (bool): Whether to automatically fix issues.
        logger (Logger): Logger object for logging purposes.
        translator (Translator): Translator object for translation purposes.
        collector (MayaVirusCollector): MayaVirusCollector object for collecting issues.
        virus_cleaner (MayaVirusCleaner): MayaVirusCleaner object for fixing issues.
        hooks (list): List of hooks to run.
    """
    _vaccines = []
    callback_maps = {
        "after_open": om.MSceneMessage.kAfterOpen,
        "maya_initialized": om.MSceneMessage.kMayaInitialized,
        "after_import": om.MSceneMessage.kAfterImport,
        "after_import_reference": om.MSceneMessage.kAfterImportReference,
        "after_load_reference": om.MSceneMessage.kAfterLoadReference,
        "before_save": om.MSceneMessage.kBeforeSave,
        "before_import": om.MSceneMessage.kBeforeImport,
        "before_load_reference": om.MSceneMessage.kBeforeLoadReference,
        "before_import_reference": om.MSceneMessage.kBeforeImportReference,
        "maya_exiting": om.MSceneMessage.kMayaExiting,
    }

    def __init__(self, auto_fix=True):
        """Initialize the MayaVirusDefender.

        Args:
            auto_fix (bool): Whether to automatically fix issues.
        """
        logger = logging.getLogger(__name__)
        self.auto_fix = auto_fix
        self.logger = setup_logger(logger)
        self.translator = Translator()
        self.collector = MayaVirusCollector(self.logger, self.translator)
        self.virus_cleaner = MayaVirusCleaner(self.collector, self.logger)
        self.hooks = get_hooks()

    def run_hooks(self):
        """Run all hooks, only works in non-batch mode."""
        if not is_maya_standalone():
            for hook_file in self.hooks:
                self.logger.debug("run_hook: %s", hook_file)
                try:
                    load_hook(hook_file).hook(virus_cleaner=self.virus_cleaner)
                except Exception as e:
                    self.logger.debug("Error running hook: %s", e)

    def collect(self):
        """Collect all issues related to the Maya virus."""
        self.collector.collect()

    def fix(self):
        """Fix all issues related to the Maya virus."""
        self.virus_cleaner.fix()

    def report(self):
        """Report all issues related to the Maya virus."""
        self.collect()
        self.collector.report()

    @property
    def have_issues(self):
        """Check if any issues are found.

        Returns:
            bool: True if any issues are found, False otherwise.
        """
        return self.collector.have_issues

    def setup(self):
        """Set up the MayaVirusDefender."""
        self.virus_cleaner.setup_default_callbacks()
        for name, callbacks in self.collector.registered_callbacks.items():
            maya_callback = self.callback_maps[name]
            self.logger.debug("%s setup.", name)
            for func in callbacks:
                _add_callbacks_id(om.MSceneMessage.addCallback(maya_callback, func))
        for name, callbacks in self.callback_maps.items():
            self.logger.debug("setup callback %s.", name)
            _add_callbacks_id(om.MSceneMessage.addCallback(callbacks, self._callback))

    def stop(self):
        """Stop the MayaVirusDefender."""
        for ids in MAYA_UMBRELLA_CALLBACK_IDS:
            self.logger.debug("remove callback. %s", ids)
            om.MSceneMessage.removeCallback(ids)
            MAYA_UMBRELLA_CALLBACK_IDS.remove(ids)

    def get_unfixed_references(self):
        """Get the list of unfixed reference files.

        Returns:
            list: List of unfixed reference files.
        """
        self.collect()
        return self.collector.infected_reference_files

    def _callback(self, *args, **kwargs):
        """Callback function for MayaVirusDefender.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        if self.auto_fix:
            self.collect()
            self.fix()
            self.run_hooks()
        else:
            self.report()

    def start(self):
        """Start the MayaVirusDefender."""
        self._callback()


@contextmanager
def context_defender():
    """Context manager for MayaVirusDefender.

    Yields:
        MayaVirusDefender: An instance of MayaVirusDefender.
    """
    defender = MayaVirusDefender()
    defender.stop()
    yield defender
    defender.setup()



# Import built-in modules
from contextlib import contextmanager
import glob
import importlib
import json
import os
import random
import re
import shutil
import string
import sys
import tempfile

# Import local modules
from maya_umbrella.constants import FILE_VIRUS_SIGNATURES
from maya_umbrella.constants import PACKAGE_NAME


PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3


def this_root():
    """Return the absolute path of the current file's directory."""
    return os.path.abspath(os.path.dirname(__file__))


def safe_remove_file(file_path):
    """Remove the file at the given path without raising an error if the file does not exist."""
    try:
        os.remove(file_path)
    except (OSError, IOError):  # noqa: UP024
        pass


def safe_rmtree(path):
    """Remove the directory at the given path without raising an error if the directory does not exist."""
    try:
        shutil.rmtree(path)
    except (OSError, IOError):  # noqa: UP024
        pass


def read_file(path):
    """Read the content of the file at the given path."""
    options = {"encoding": "utf-8"} if PY3 else {}
    with open(path, **options) as file_:
        try:
            content = file_.read()
        # maya-2022 UnicodeDecodeError from `plug-ins/mayaHIK.pres.mel`
        except UnicodeDecodeError:
            return ""
    return content


def read_json(path):
    """Read the content of the file at the given path."""
    options = {"encoding": "utf-8"} if PY3 else {}
    with open(path, **options) as file_:
        try:
            content = json.load(file_)
        except UnicodeDecodeError:
            return {}
    return content


def write_file(path, content):
    """Write the given content to the file at the given path."""
    options = {"encoding": "utf-8"} if PY3 else {}
    with atomic_writes(path, "w", **options) as file_:
        file_.write(content)


@contextmanager
def atomic_writes(src, mode, **options):
    """Context manager for atomic writes to a file.

    This context manager ensures that the file is only written to disk if the write operation completes without errors.

    Args:
        src (str): Path to the file to be written.
        mode (str): Mode in which the file is opened, like 'r', 'w', 'a', etc.
        **options: Arbitrary keyword arguments that are passed to the built-in open() function.

    Yields:
        file object: The opened file object.

    Raises:
        AttributeError: If the os module does not have the 'replace' function (Python 2 compatibility).
    """
    temp_path = os.path.join(os.path.dirname(src), "._{}".format(id_generator()))
    with open(temp_path, mode, **options) as f:
        yield f
    try:
        os.replace(temp_path, src)
    except AttributeError:
        shutil.move(temp_path, src)



def id_generator(size=6, chars=string.ascii_uppercase + string.digits):
    """Generate a random string of the given size using the given characters."""
    return "".join(random.choice(chars) for _ in range(size))


def rename(src):
    """Rename the file at the given path to a random name and return the new path."""
    dst = os.path.join(os.path.dirname(src), "._{}".format(id_generator()))
    try:
        os.rename(src, dst)
    except (OSError, IOError):  # noqa: UP024
        return src
    return dst


def load_hook(hook_file):
    """Load the Python module from the given hook file."""
    hook_name = os.path.basename(hook_file).split(".py")[0]
    if hasattr(importlib, "machinery"):
        # Python 3
        # Import built-in modules
        from importlib.util import spec_from_loader  # noqa: F401

        loader = importlib.machinery.SourceFileLoader(hook_name, hook_file)
        spec = importlib.util.spec_from_loader(loader.name, loader=loader)
        module = importlib.util.module_from_spec(spec)
        loader.exec_module(module)
    else:
        # Python 2
        # Import built-in modules
        import imp

        module = imp.load_source(hook_name, hook_file)
    return module


def get_hooks():
    """Return a list of paths to all hook files in the 'hooks' directory."""
    pattern = os.path.join(this_root(), "hooks", "*.py")
    return [hook for hook in glob.glob(pattern) if "__init__" not in hook]


def get_vaccines():
    """Get a list of all vaccine files.

    Returns:
        list: A list of vaccine files.
    """
    pattern = os.path.join(this_root(), "vaccines", "*.py")
    return [vaccine for vaccine in glob.glob(pattern) if "__init__" not in vaccine]


def get_log_root():
    """Get the log root directory.

    Returns:
        str: The log root directory.
    """
    return os.getenv("MAYA_UMBRELLA_LOG_ROOT", tempfile.gettempdir())


def get_log_file():
    """Get the path of the log file.

    Returns:
        str: The path of the log file.
    """
    root = get_log_root()
    try:
        os.makedirs(root)
    except (OSError, IOError):  # noqa: UP024
        pass
    name = os.getenv("MAYA_UMBRELLA_LOG_NAME", PACKAGE_NAME)
    return os.path.join(root, "{name}.log".format(name=name))


def remove_virus_file_by_signature(file_path, signatures, output_file_path=None):
    """Remove virus content from a file by matching signatures.

    Args:
        file_path (str): Path to the file to be cleaned.
        signatures (list): List of signatures to match and remove.
        output_file_path (str, optional): Path to the cleaned output file.
         Defaults to None, which overwrites the input file.
    """
    data = read_file(file_path)
    if check_virus_by_signature(data, signatures):
        fixed_data = replace_content_by_signatures(data, signatures)
        write_file(output_file_path or file_path, fixed_data)


def replace_content_by_signatures(content, signatures):
    """Replace content in a string that matches given signatures.

    Args:
        content (str): The input content.
        signatures (list): List of signatures to match and remove.

    Returns:
        str: The cleaned content.
    """
    for signature in signatures:
        content = re.sub(signature, "", content)
    return content


def check_virus_file_by_signature(file_path, signatures=None):
    """Check if a file contains a virus by matching signatures.

    Args:
        file_path (str): Path to the file to be checked.
        signatures (list, optional): List of signatures to match. Defaults to None, which uses FILE_VIRUS_SIGNATURES.

    Returns:
        bool: True if a virus signature is found, False otherwise.
    """
    signatures = signatures or FILE_VIRUS_SIGNATURES
    try:
        data = read_file(file_path)
        return check_virus_by_signature(data, signatures)
    except (OSError, IOError):  # noqa: UP024
        return False
    except UnicodeDecodeError:
        return True


def check_virus_by_signature(content, signatures=None):
    """Check if a content contains a virus by matching signatures.

    Args:
        content (str): The input content.
        signatures (list, optional): List of signatures to match. Defaults to None, which uses FILE_VIRUS_SIGNATURES.

    Returns:
        bool: True if a virus signature is found, False otherwise.
    """
    signatures = signatures or FILE_VIRUS_SIGNATURES
    for signature in signatures:
        if re.search(signature, content):
            return True
    return False


def get_backup_path(path, root_path=None):
    """Get the backup path for a given file path based on environment variables.

    Args:
        path (str): Path to the original file.
        root_path (str, optional): Path to the root folder where backups should be saved.
            Defaults to None, which saves backups in the original file's folder.

    Returns:
        str: The backup path.
    """
    ignore_backup = os.getenv("MAYA_UMBRELLA_IGNORE_BACKUP", "false").lower() == "true"
    if ignore_backup:
        return path
    root, filename = os.path.split(path)
    backup_folder_name = os.getenv("MAYA_UMBRELLA_BACKUP_FOLDER_NAME", "_virus")
    backup_path = os.path.join(root, backup_folder_name)
    if root_path:
        _, base_path = os.path.splitdrive(root)
        backup_path = os.path.join(root_path, base_path.strip(os.sep))
    try:
        os.makedirs(backup_path)
    except (OSError, IOError):  # noqa: UP024
        pass
    return os.path.join(backup_path, filename)



# Import built-in modules
import glob
import logging
import os
import shutil

# Import local modules
from maya_umbrella import maya_funs
from maya_umbrella.defender import context_defender
from maya_umbrella.filesystem import get_backup_path
from maya_umbrella.filesystem import read_file
from maya_umbrella.maya_funs import cmds


class MayaVirusScanner(object):
    """A class to scan and fix Maya files containing viruses.

    Attributes:
        _failed_files (list): List of files that failed to be fixed.
        _fixed_files (list): List of files that have been fixed.
        logger (Logger): Logger object for logging purposes.
        defender (MayaVirusDefender): MayaVirusDefender object for fixing issues.
        _env (dict): Custom environment variables.
        output_path (str, optional): Path to save the fixed files. Defaults to None, which overwrites the original
            files.
    """

    def __init__(self, output_path=None, env=None):
        """Initialize the MayaVirusScanner.

        Args:
            output_path (str, optional): Path to save the fixed files. Defaults to None, which overwrites the original
            files.
            env (dict, optional): Custom environment variables. Defaults to None,
            which sets the 'MAYA_COLOR_MANAGEMENT_SYNCOLOR' variable to '1'.
        """
        self.logger = logging.getLogger(__name__)
        self.defender = None
        self.output_path = output_path
        self._failed_files = []
        self._reference_files = []
        self._fixed_files = []
        # Custom env.
        self._env = env or {
            "MAYA_COLOR_MANAGEMENT_SYNCOLOR": "1"
        }

    def scan_files_from_pattern(self, pattern):
        """Scan and fix Maya files matching a given pattern.

        Args:
            pattern (str): The file pattern to match.
        """
        os.environ.update(self._env)
        return self.scan_files_from_list(glob.iglob(pattern))

    def scan_files_from_list(self, files):
        """Scan and fix Maya files from a given list.

        Args:
            files (list): List of file paths to scan and fix.
        """
        with context_defender() as defender:
            self.defender = defender
            for maya_file in files:
                self._fix(maya_file)
                for ref in self._reference_files:
                    self._fix(ref)
        return self._fixed_files

    def scan_files_from_file(self, text_file):
        """Scan and fix Maya files from a given text file containing a list of file paths.

        Args:
            text_file (str): Path to the text file containing the list of file paths.
        """
        file_data = read_file(text_file)
        files = file_data.splitlines()
        return self.scan_files_from_list(files)

    def _fix(self, maya_file):
        """Fix a single Maya file containing a virus.

        Args:
            maya_file (str): Path to the Maya file to be fixed.
        """
        if not maya_file and maya_file in self._fixed_files:
            self.logger.debug("Already fixed: {maya_file}".format(maya_file=maya_file))
            return
        try:
            maya_funs.open_maya_file(maya_file)
            self.defender.collect()
        except Exception:
            self.logger.debug("failed to open maya file: {maya_file}".format(maya_file=maya_file))
            self._failed_files.append(maya_file)

        if self.defender.have_issues:
            self.defender.fix()
            backup_path = get_backup_path(maya_file, root_path=self.output_path)
            self.logger.debug("Backup saved to: {backup_path}".format(backup_path=backup_path))
            shutil.copy2(maya_file, backup_path)
            cmds.file(s=True, f=True)
            self._fixed_files.append(maya_file)
            self._reference_files.extend(self.defender.collector.infected_reference_files)
        cmds.file(new=True, force=True)



# Import built-in modules
import glob
import os
from string import Template

# Import local modules
from maya_umbrella.filesystem import read_json
from maya_umbrella.filesystem import this_root
from maya_umbrella.maya_funs import maya_ui_language


class Translator(object):
    """A class to handle translations for different locales.

    Attributes:
        data (dict): Dictionary containing translation data for different locales.
        locale (str): The current locale.
    """
    def __init__(self, file_format="json", default_locale=None):
        """Initialize the Translator.

        Args:
            file_format (str, optional): File format of the translation files. Defaults to "json".
            default_locale (str, optional): Default locale to use for translations. Defaults to None,
            which uses the MAYA_UMBRELLA_LANG environment variable or the Maya UI language.
        """
        _default_locale = os.getenv("MAYA_UMBRELLA_LANG", maya_ui_language())
        default_locale = default_locale or _default_locale
        self.data = {}
        self.locale = default_locale
        translations_folder = os.path.join(this_root(), "locales")

        # get list of files with specific extensions
        files = glob.glob(os.path.join(translations_folder, "*.{file_format}".format(file_format=file_format)))
        for fil in files:
            # get the name of the file without extension, will be used as locale name
            loc = os.path.splitext(os.path.basename(fil))[0]
            self.data[loc] = read_json(fil)

    def set_locale(self, locale):
        """Set the current locale.

        Args:
            locale (str): The locale to set.

        Raises:
            ValueError: If the provided locale is not supported.
        """
        if locale in self.data:
            self.locale = locale
        else:
            raise ValueError("Invalid locale: {loc}".format(loc=locale))

    def get_locale(self):
        """Get the current locale.

        Returns:
            str: The current locale.
        """
        return self.locale

    def translate(self, key, **kwargs):
        """Translate a text based on the current locale.

        Args:
            key (str): The key to be translated.
            **kwargs: Arbitrary keyword arguments that are used to replace placeholders in the translation text.

        Returns:
            str: The translated text.
        """
        # return the key instead of translation text if locale is not supported
        if self.locale not in self.data:
            return key
        text = self.data[self.locale].get(key, key)
        return Template(text).safe_substitute(**kwargs)



PACKAGE_NAME = "maya_umbrella"

LOG_FORMAT = "%(asctime)s [%(levelname)s] %(name)s: %(message)s"

LOG_MAX_BYTES = 1024 * 1024 * 5

FILE_VIRUS_SIGNATURES = [
    "import vaccine",
    "cmds.evalDeferred.*leukocyte.+",
    # https://regex101.com/r/0MNzF7/1
    "python(.*);.+exec.+(pyCode).+;",
]

JOB_SCRIPTS_VIRUS_SIGNATURES = [
    "petri_dish_path.+cmds.internalVar.+",
    "userSetup",
    "fuckVirus",
    # https://regex101.com/r/0MNzF7/1
    "python(.*);.+exec.+(pyCode).+;",
    # https://regex101.com/r/2D14UA/1
    r"^\['.+']",
]



"""In order to facilitate unit testing in CI.

we extracted all the interfaces used in Maya and mocked them uniformly.

"""

# Import third-party modules
try:
    # Import third-party modules
    import maya.api.OpenMaya as om
    import maya.cmds as cmds
    import maya.mel as mel
except ImportError:
    # Backward compatibility to support test in uinstalled maya.
    try:
        # Import built-in modules
        from unittest.mock import MagicMock
    except ImportError:
        # Import third-party modules
        from mock import MagicMock  # noqa: UP026
    cmds = MagicMock()
    om = MagicMock()
    mel = MagicMock()

# Import built-in modules
from functools import wraps


def is_maya_standalone():
    """Check if Maya is running in standalone mode.

    Returns:
        bool: True if Maya is running in standalone mode, False otherwise.
    """
    return cmds.about(batch=True)


def check_reference_node_exists(node_name):
    """Check if a reference node exists in the Maya scene.

    Args:
        node_name (str): Name of the reference node.

    Returns:
        bool: True if the reference node exists, False otherwise.
    """
    try:
        return cmds.referenceQuery(node_name, isNodeReferenced=True)
    except RuntimeError:
        return False


def get_reference_file_by_node(node_name):
    """Get the reference file associated with a node.

    Args:
        node_name (str): Name of the node.

    Returns:
        str: Path of the reference file, empty string if the node is not associated with a reference file.
    """
    try:
        return cmds.referenceQuery(node_name, filename=True)
    except RuntimeError:
        return ""


def get_attr_value(node_name, attr_name):
    """Get the value of an attribute of a node.

    Args:
        node_name (str): Name of the node.
        attr_name (str): Name of the attribute.

    Returns:
        Any: Value of the attribute, None if the attribute does not exist.
    """
    try:
        return cmds.getAttr("{node_name}.{attr}".format(node_name=node_name, attr=attr_name))
    except ValueError:
        return None


def maya_ui_language():
    """Get the language of the Maya user interface.

    Returns:
        str: The language of the Maya user interface.
    """
    return cmds.about(uiLocaleLanguage=True)


def block_prompt(func):
    """Decorator to block file prompt dialogs in Maya.

    Args:
        func (function): The function to decorate.

    Returns:
        function: The decorated function.
    """
    @wraps(func)
    def wrap(*args, **kwargs):
        # Grabs the initial value.
        prompt_val = cmds.file(prompt=True, q=True)

        try:
            cmds.file(prompt=False)
            return func(*args, **kwargs)

        finally:
            # Resets to the original value, this way you don't suddenly turn the prompt on, when someone wanted it off.
            cmds.file(prompt=prompt_val)

    return wrap


@block_prompt
def open_maya_file(maya_file):
    """Open a Maya file.

    Args:
        maya_file (str): Path to the Maya file.
    """
    cmds.file(maya_file, open=True, force=True, ignoreVersion=True, executeScriptNodes=False)


@block_prompt
def save_as_file(file_name):
    """Save the current Maya scene as a file.

    Args:
        file_name (str): Path to the output file.
    """
    cmds.file(rename=file_name)
    cmds.file(s=True, f=True)



# Import local modules
from maya_umbrella.maya_funs import cmds


def hook(virus_cleaner):
    for model_panel in cmds.getPanel(typ="modelPanel"):
        # Get callback of the model editor
        callback = cmds.modelEditor(model_panel, query=True, editorChanged=True)

        # If the callback is the erroneous `CgAbBlastPanelOptChangeCallback`
        if callback == "CgAbBlastPanelOptChangeCallback":
            virus_cleaner.logger.info("Remove the callbacks from the editor")
            cmds.modelEditor(model_panel, edit=True, editorChanged="")



# Import local modules
from maya_umbrella.maya_funs import cmds
from maya_umbrella.maya_funs import mel


def hook(virus_cleaner):
    try:
        mel.eval("global proc onModelChange3dc(string $a){}")
    except Exception:
        pass
    try:
        cmds.delete("fixCgAbBlastPanelOptChangeCallback")
    except Exception:
        pass
    try:
        script = "global proc CgAbBlastPanelOptChangeCallback(string $i){}"
        mel.eval(script)
    except Exception:
        pass



# Import built-in modules
import logging.handlers
import os

# Import local modules
from maya_umbrella.constants import LOG_FORMAT
from maya_umbrella.constants import LOG_MAX_BYTES
from maya_umbrella.constants import PACKAGE_NAME
from maya_umbrella.filesystem import get_log_file


def setup_logger(logger=None, logfile=None, log_level=None):
    """Set up the logger with the specified log file and log level.

    Args:
        logger (logging.Logger, optional): The logger to set up. Defaults to the logger for the package.
        logfile (str, optional): The path to the log file. Defaults to the log file returned by `get_log_file()`.
        log_level (int, optional): The log level. Defaults to `logging.INFO`.

    Returns:
        logging.Logger: The set up logger.
    """
    logger = logger or logging.getLogger(PACKAGE_NAME)
    log_level = log_level or os.getenv("MAYA_UMBRELLA_LOG_LEVEL", "INFO")
    logger.setLevel(log_level)
    logfile = logfile or get_log_file()
    if not len(logger.handlers):
        filehandler = logging.handlers.RotatingFileHandler(
            logfile,
            mode="a",
            backupCount=7,
            delay=True,
            maxBytes=LOG_MAX_BYTES,
        )
        filehandler.setFormatter(logging.Formatter(LOG_FORMAT))
        logger.addHandler(filehandler)
    return logger



class AbstractVaccine(object):
    """Abstract base class for Vaccine classes.

    Attributes:
        virus_name (str): The name of the virus.
        api (MayaVirusCleaner): The VaccineAPI instance.
        logger (Logger): The logger instance.
    """

    virus_name = None

    def __init__(self, api, logger):
        """Abstract class for Vaccine.

        Args:
            api (MayaVirusCleaner): The VaccineAPI instance.
            logger (Logger): The logger instance.

        """
        self.api = api
        self.logger = logger

    def collect_issues(self):
        """Collect issues related to the virus.

        Raises:
            NotImplementedError: This method must be implemented in the derived classes.
        """
        raise NotImplementedError

    def report_issue(self, name):
        """Report an issue related to the virus.

        Args:
            name (str): The name of the issue.
        """
        self.logger.warning(self.api.translator.translate("report_issue", name=name))



# Import built-in modules
import os.path

# Import local modules
from maya_umbrella.constants import JOB_SCRIPTS_VIRUS_SIGNATURES
from maya_umbrella.filesystem import check_virus_by_signature
from maya_umbrella.filesystem import check_virus_file_by_signature
from maya_umbrella.maya_funs import check_reference_node_exists
from maya_umbrella.maya_funs import cmds
from maya_umbrella.maya_funs import get_attr_value
from maya_umbrella.vaccine import AbstractVaccine


class Vaccine(AbstractVaccine):
    """A class for handling the ZeiJianKang virus."""

    virus_name = "zei jian kang"

    def collect_infected_nodes(self):
        """Collect all bad nodes related to the virus."""
        for script_node in cmds.ls(type="script"):
            if check_reference_node_exists(script_node):
                continue
            for attr_name in ("before", "after"):
                script_string = get_attr_value(script_node, attr_name)
                if not script_string:
                    continue
                if check_virus_by_signature(script_string, JOB_SCRIPTS_VIRUS_SIGNATURES):
                    self.report_issue(script_node)
                    self.api.add_infected_node(script_node)

    def collect_issues(self):
        """Collect all issues related to the virus."""
        self.api.add_malicious_files(
            [
                os.path.join(self.api.local_script_path, "vaccine.py"),
                os.path.join(self.api.local_script_path, "vaccine.pyc"),
            ],
        )
        self.collect_infected_user_setup_py()
        self.collect_infected_nodes()

    def collect_infected_user_setup_py(self):
        """Collect all bad userSetup.py files related to the virus."""
        for user_setup_py in [
            os.path.join(self.api.local_script_path, "userSetup.py"),
            os.path.join(self.api.user_script_path, "userSetup.py"),
        ]:
            if os.path.exists(user_setup_py):
                if check_virus_file_by_signature(user_setup_py):
                    self.report_issue(user_setup_py)
                    self.api.add_infected_file(user_setup_py)



# Import built-in modules
import glob
import os

# Import local modules
from maya_umbrella.constants import JOB_SCRIPTS_VIRUS_SIGNATURES
from maya_umbrella.filesystem import check_virus_by_signature
from maya_umbrella.filesystem import check_virus_file_by_signature
from maya_umbrella.maya_funs import cmds
from maya_umbrella.maya_funs import get_attr_value
from maya_umbrella.maya_funs import get_reference_file_by_node
from maya_umbrella.maya_funs import is_maya_standalone
from maya_umbrella.vaccine import AbstractVaccine


class Vaccine(AbstractVaccine):
    """A class for handling the virus2024429 virus."""

    virus_name = "Virus2024429"

    def collect_infected_nodes(self):
        """Collect all bad nodes related to the virus."""
        for script_node in cmds.ls(type="script"):
            # check vaccine
            if "_gene" in script_node:
                self.report_issue(script_node)
                self.api.add_infected_node(script_node)
                self.api.add_infected_reference_file(get_reference_file_by_node(script_node))
            if "uifiguration" in script_node:
                for attr_name in ("before", "notes"):
                    script_string = get_attr_value(script_node, attr_name)
                    if not script_string:
                        continue
                if check_virus_by_signature(script_string, JOB_SCRIPTS_VIRUS_SIGNATURES):
                    self.report_issue(script_node)
                    self.api.add_infected_node(script_node)
                    self.api.add_infected_reference_file(get_reference_file_by_node(script_node))

    def collect_infected_mel_files(self):
        """Collect all bad MEL files related to the virus."""
        # check usersetup.mel
        # C:/Users/hallong/Documents/maya/scripts/usersetup.mel
        # C:/Users/hallong/Documents/maya/xxxx/scripts/usersetup.mel
        for usersetup_mel in [
            os.path.join(self.api.local_script_path, "usersetup.mel"),
            os.path.join(self.api.user_script_path, "usersetup.mel"),
        ]:
            if os.path.exists(usersetup_mel):
                if check_virus_file_by_signature(usersetup_mel):
                    self.report_issue(usersetup_mel)
                    self.api.add_infected_file(usersetup_mel)

    def collect_script_jobs(self):
        """Collect all script jobs related to the virus."""
        virus_gene = [
            "leukocyte",
            "execute",
        ]
        for script_job in cmds.scriptJob(listJobs=True):
            for virus in virus_gene:
                if virus in script_job:
                    self.api.add_infected_script_job(script_job)

    def fix_bad_hik_files(self):
        """Fix all bad HIK files related to the virus."""
        pattern = os.path.join(self.api.maya_install_root, "resources/l10n/*/plug-ins/mayaHIK.pres.mel")
        for hik_mel in glob.glob(pattern):
            if check_virus_file_by_signature(hik_mel):
                self.report_issue(hik_mel)
                self.api.add_infected_file(hik_mel)

    def collect_issues(self):
        """Collect all issues related to the virus."""
        self.api.add_malicious_file(os.path.join(os.getenv("APPDATA"), "syssst"))
        self.collect_infected_mel_files()
        self.collect_infected_nodes()
        # This only works for Maya Gui model.
        if not is_maya_standalone():
            self.collect_script_jobs()
        self.api.add_additionally_fix_function(self.fix_bad_hik_files)



# Import local modules
from maya_umbrella.cleaner import MayaVirusCleaner
from maya_umbrella.collector import MayaVirusCollector
from maya_umbrella.defender import MayaVirusDefender
from maya_umbrella.defender import context_defender
from maya_umbrella.scanner import MayaVirusScanner


__all__ = ["MayaVirusDefender", "MayaVirusCleaner", "MayaVirusCollector", "MayaVirusScanner", "context_defender"]



# Import built-in modules
import os

# Import local modules
from maya_umbrella.vaccine import AbstractVaccine


class Vaccine(AbstractVaccine):
    """A class for handling the PuTianTongQi virus."""
    virus_name = "PutTianTongQi"

    def collect_issues(self):
        self.api.add_malicious_files(
            [
                os.path.join(self.api.local_script_path, "fuckVirus.py"),
                os.path.join(self.api.local_script_path, "fuckVirus.pyc"),
            ],
        )



# Import local modules
from maya_umbrella.maya_funs import cmds
from maya_umbrella.maya_funs import mel


def hook(virus_cleaner):
    for plugin in ["Turtle.mll", "mayatomr.mll"]:
        if cmds.pluginInfo(plugin, q=1, loaded=1):
            cmds.unloadPlugin(plugin, f=1)
    turtle_nodes = [
        "TurtleRenderOptions",
        "TurtleUIOptions",
        "TurtleBakeLayerManager",
        "TurtleDefaultBakeLayer",
    ]
    for node in turtle_nodes:
        if cmds.objExists(node):
            cmds.lockNode(node, lock=1)
            cmds.delete(node)
    if not cmds.about(query=1, batch=1):
        shelves = cmds.tabLayout(mel.eval("$tmpVar=$gShelfTopLevel"), q=1, ca=1)
        if "TURTLE" in shelves:
            cmds.deleteUI("TURTLE", layout=1)



# Import built-in modules
import glob
import logging
import os
import re

# Import local modules
from maya_umbrella.constants import FILE_VIRUS_SIGNATURES
from maya_umbrella.filesystem import remove_virus_file_by_signature
from maya_umbrella.filesystem import safe_remove_file
from maya_umbrella.filesystem import safe_rmtree
from maya_umbrella.i18n import Translator
from maya_umbrella.maya_funs import check_reference_node_exists
from maya_umbrella.maya_funs import cmds


class MayaVirusCleaner(object):
    """A class to clean Maya virus files.

    Attributes:
        logger (Logger): Logger object for logging purposes.
        translator (Translator): Translator object for translation purposes.
        collector (MayaVirusCollector): MayaVirusCollector object for collecting issues.
    """
    def __init__(self, collector, logger=None):
        """Initialize the MayaVirusCleaner.

        Args:
            collector (MayaVirusCollector): MayaVirusCollector object for collecting issues.
            logger (Logger, optional): Logger object for logging purposes. Defaults to None, which creates a new logger.
        """
        self.logger = logger or logging.getLogger(__name__)
        self.translator = Translator()
        self.collector = collector

    def callback_remove_rename_temp_files(self, *args, **kwargs):
        """Remove temporary files in the local script path."""
        self.logger.debug("Removing temporary files in %s", self.collector.local_script_path)
        for temp_file in glob.glob(os.path.join(self.collector.local_script_path, "._*")):
            safe_remove_file(temp_file)

    def fix_script_jobs(self):
        """Fix infected script jobs."""
        for script_job in self.collector.infected_script_jobs:
            script_num = int(re.findall(r"^(\d+):", script_job)[0])
            self.logger.debug("Kill script job %s", script_job)
            cmds.scriptJob(kill=script_num, force=True)
            self.collector.remove_infected_script_job(script_job)

    def fix_malicious_files(self):
        """Fix malicious files."""
        for file_ in self.collector.malicious_files:
            if os.path.exists(file_):
                if os.path.isfile(file_):
                    self.logger.debug(self.translator.translate("remove_file", name=file_))
                    safe_remove_file(file_)
                    self.collector.remove_malicious_file(file_)
                else:
                    self.logger.debug(self.translator.translate("remove_path", name=file_))
                    safe_rmtree(file_)
                    self.collector.remove_malicious_file(file_)

    def fix_infected_nodes(self):
        """Fix infected nodes."""
        for node in self.collector.infected_nodes:
            is_referenced = check_reference_node_exists(node)
            if is_referenced:
                try:
                    self.logger.debug(self.translator.translate("fix_infected_reference_nodes", name=node))
                    cmds.setAttr("{node}.before".format(node=node), "", type="string")
                    cmds.setAttr("{node}.after".format(node=node), "", type="string")
                    cmds.setAttr("{node}.scriptType".format(node=node), 0)
                    self.collector.remove_infected_node(node)
                except Exception as e:
                    self.logger.debug(e)
            else:
                try:
                    cmds.lockNode(node, lock=False)
                except ValueError:
                    pass
                try:
                    self.logger.debug(self.translator.translate("fix_infected_nodes", name=node))
                    cmds.delete(node)
                except ValueError:
                    pass
                self.collector.remove_infected_node(node)

    def setup_default_callbacks(self):
        """Set up default callbacks."""
        self.collector.add_maya_initialized_callback(self.callback_remove_rename_temp_files)
        self.collector.add_maya_exiting_callback(self.callback_remove_rename_temp_files)

    def fix_infected_files(self):
        """Fix infected files."""
        for file_path in self.collector.infected_files:
            self.logger.info(self.translator.translate("fix_infected_files", name=file_path))
            remove_virus_file_by_signature(file_path, FILE_VIRUS_SIGNATURES)
            self.collector.remove_infected_file(file_path)

    def fix(self):
        """Fix all issues related to the Maya virus."""
        if self.collector.have_issues:
            maya_file = cmds.file(query=True, sceneName=True, shortName=True) or "empty/scene"
            self.logger.info(self.translator.translate("start_fix_issues", name=maya_file))
            self.fix_malicious_files()
            self.fix_infected_files()
            self.fix_infected_nodes()
            self.fix_script_jobs()
            for func in self.collector.get_additionally_fix_funcs():
                func()
            self.logger.info(self.translator.translate("finish_fix_issues", name=maya_file))



# Import local modules
from maya_umbrella.maya_funs import cmds


def hook(virus_cleaner):
    unknown_node = cmds.ls(type="unknown")
    unknown_plugin = cmds.unknownPlugin(query=True, l=True)
    if unknown_node:
        for nodeObj in unknown_node:
            if cmds.objExists(nodeObj):
                if cmds.referenceQuery(nodeObj, isNodeReferenced=True):
                    virus_cleaner.logger.warning("Node from reference, skip. {}".format(nodeObj))
                    continue
                if cmds.lockNode(nodeObj, query=True)[0]:
                    try:
                        cmds.lockNode(nodeObj, lock=False)
                    except Exception:
                        virus_cleaner.logger.warning(
                            "The node is locked and cannot be unlocked. skip {}".format(nodeObj)
                        )
                        continue
                try:
                    cmds.delete(nodeObj)
                    virus_cleaner.logger.warning("Delete node: {}".format(nodeObj))
                except Exception:
                    pass

    if unknown_plugin:
        for plugObj in unknown_plugin:
            cmds.unknownPlugin(plugObj, remove=True)
            virus_cleaner.logger.warning("Delete plug-in: {}".format(plugObj))



# Import built-in modules
from collections import defaultdict
import logging
import os

# Import local modules
from maya_umbrella.filesystem import get_vaccines
from maya_umbrella.filesystem import load_hook
from maya_umbrella.i18n import Translator
from maya_umbrella.maya_funs import cmds


class MayaVirusCollector(object):
    """A class to collect and handle Maya viruses.

    Attributes:
        _malicious_files (list): List to store malicious files.
        _infected_files (list): List to store infected files.
        _infected_nodes (list): List to store infected nodes.
        _infected_script_nodes (list): List to store infected script nodes.
        _infected_reference_files (list): List to store infected reference files.
        _infected_script_jobs (list): List to store infected script jobs.
        _registered_callbacks (defaultdict): Dictionary to store registered callbacks.
        _additionally_fix_funcs (list): List to store additional fix functions.
        _vaccines (list): List to store vaccines.
        logger: Logger object for logging purposes.
        translator: Translator object for translation purposes.
    """

    def __init__(self, logger, translator=None):
        """Initialize MayaVirusCollector.

        Args:
            logger (Logger, optional): Logger object for logging purposes.
            translator (Translator, optional): Translator object for translation purposes.
        """
        self.logger = logger or logging.getLogger(__name__)
        self.translator = translator or Translator()
        # Malicious files or temp files that need to be deleted directly.
        self._malicious_files = []
        self._infected_files = []
        self._infected_nodes = []
        self._infected_script_nodes = []
        self._infected_reference_files = []
        self._infected_script_jobs = []
        self._registered_callbacks = defaultdict(list)
        self._additionally_fix_funcs = []
        self._vaccines = []
        self.load_vaccines()

    def load_vaccines(self):
        """Load all vaccines."""
        for vaccine in get_vaccines():
            vaccine_class = load_hook(vaccine).Vaccine
            try:
                self._vaccines.append(vaccine_class(api=self, logger=self.logger))
            except Exception as e:
                self.logger.error("Error loading vaccine: %s", e)

    @property
    def vaccines(self):
        """Get all loaded vaccines.

        Returns:
            list: A list of loaded vaccines.
        """
        return self._vaccines

    @property
    def user_app_dir(self):
        """Return the user application directory."""
        return cmds.internalVar(userAppDir=True)

    @property
    def maya_install_root(self):
        """Return the Maya installation root directory."""
        return os.environ["MAYA_LOCATION"]

    @property
    def user_script_path(self):
        """Return the user script directory."""
        return cmds.internalVar(userScriptDir=True)

    @property
    def local_script_path(self):
        """Return the local script directory."""
        return os.path.join(self.user_app_dir, "scripts")

    @property
    def malicious_files(self):
        """Return a list of bad files."""
        return [path for path in list(set(self._malicious_files)) if os.path.exists(path)]

    @property
    def infected_nodes(self):
        """Return a list of infected nodes."""
        return list(set(self._infected_nodes))

    @property
    def infected_reference_files(self):
        return [path for path in list(set(self._infected_reference_files)) if os.path.exists(path)]

    @property
    def infected_script_nodes(self):
        """Return a list of bad script nodes."""
        return list(set(self._infected_script_nodes))

    @property
    def infected_script_jobs(self):
        """Return a list of bad script jobs."""
        return list(set(self._infected_script_jobs))

    @property
    def infected_files(self):
        """Return the list of infected files.

        Returns:
            list: List of infected files.
        """
        return self._infected_files

    @property
    def registered_callbacks(self):
        """Return the dictionary of registered callbacks.

        Returns:
            defaultdict: Dictionary of registered callbacks.
        """
        return self._registered_callbacks

    def add_infected_reference_files(self, files):
        """Add multiple infected reference files.

        Args:
            files (list): List of infected reference files to be added.
        """
        self._infected_reference_files.extend(files)

    def add_infected_reference_file(self, file):
        """Add a single infected reference file.

        Args:
            file (str): Infected reference file to be added.
        """
        self._infected_reference_files.append(file)

    def remove_infected_reference_file(self, file):
        """Remove an infected reference file.

        Args:
            file (str): Infected reference file to be removed.
        """
        self._infected_reference_files.remove(file)

    def add_infected_files(self, files):
        """Add multiple infected files.

        Args:
            files (list): List of infected files to be added.
        """
        self._infected_files.extend(files)

    def add_infected_file(self, file):
        """Add a single infected file.

        Args:
            file (str): Infected file to be added.
        """
        self._infected_files.append(file)

    def remove_infected_file(self, file):
        """Remove an infected file.

        Args:
            file (str): Infected file to be removed.
        """
        self._infected_files.remove(file)

    def add_malicious_files(self, files):
        """Add multiple malicious files.

        Args:
            files (list): List of malicious files to be added.
        """
        self._malicious_files.extend(files)

    def add_malicious_file(self, file):
        """Add a single malicious file.

        Args:
            file (str): Malicious file to be added.
        """
        self._malicious_files.append(file)

    def remove_malicious_file(self, file):
        """Remove a malicious file.

        Args:
            file (str): Malicious file to be removed.
        """
        self._malicious_files.remove(file)

    def add_infected_nodes(self, nodes):
        """Add multiple infected nodes.

        Args:
            nodes (list): List of infected nodes to be added.
        """
        self._infected_nodes.extend(nodes)

    def add_infected_node(self, node):
        """Add a single infected node.

        Args:
            node (str): Infected node to be added.
        """
        self._infected_nodes.append(node)

    def remove_infected_node(self, node):
        """Remove an infected node.

        Args:
            node (str): Infected node to be removed.
        """
        self._infected_nodes.remove(node)

    def add_infected_script_jobs(self, jobs):
        """Add multiple infected script jobs.

        Args:
            jobs (list): List of infected script jobs to be added.
        """
        self._infected_script_jobs.extend(jobs)

    def add_infected_script_job(self, job):
        """Add a single infected script job.

        Args:
            job (str): Infected script job to be added.
        """
        self._infected_script_jobs.append(job)

    def remove_infected_script_job(self, job):
        """Remove an infected script job.

        Args:
            job (str): Infected script job to be removed.
        """
        self._infected_script_jobs.remove(job)

    def add_infected_script_nodes(self, nodes):
        """Add multiple infected script nodes.

        Args:
            nodes (list): List of infected script nodes to be added.
        """
        self._infected_script_nodes.extend(nodes)

    def add_infected_script_node(self, node):
        """Add a single infected script node.

        Args:
            node (str): Infected script node to be added.
        """
        self._infected_script_nodes.append(node)

    def remove_infected_script_node(self, node):
        """Remove an infected script node.

        Args:
            node (str): Infected script node to be removed.
        """
        self._infected_script_nodes.remove(node)

    def get_additionally_fix_funcs(self):
        """Return the list of additional fix functions.

        Returns:
            list: List of additional fix functions.
        """
        return self._additionally_fix_funcs

    def register_callback(self, callback_name, callback):
        """Register a callback to be executed before or after processing."""
        self._registered_callbacks[callback_name].append(callback)

    def add_after_open_callback(self, callback):
        """Add a callback to be executed after a file is opened.

        Args:
            callback (function): The callback function to be added.
        """
        self.register_callback("after_open", callback)

    def add_maya_initialized_callback(self, callback):
        """Add a callback to be executed when Maya is initialized.

        Args:
            callback (function): The callback function to be added.
        """
        self.register_callback("maya_initialized", callback)

    def add_after_import_callback(self, callback):
        """Add a callback to be executed after a file is imported.

        Args:
            callback (function): The callback function to be added.
        """
        self.register_callback("after_import", callback)

    def add_after_import_reference_callback(self, callback):
        """Add a callback to be executed after a reference file is imported.

        Args:
            callback (function): The callback function to be added.
        """
        self.register_callback("after_import_reference", callback)

    def add_after_load_reference_callback(self, callback):
        """Add a callback to be executed after a reference file is loaded.

        Args:
            callback (function): The callback function to be added.
        """
        self.register_callback("after_load_reference", callback)

    def add_before_save_callback(self, callback):
        """Add a callback to be executed before a file is saved.

        Args:
            callback (function): The callback function to be added.
        """
        self.register_callback("before_save", callback)

    def add_before_import_callback(self, callback):
        """Add a callback to be executed before a file is imported.

        Args:
            callback (function): The callback function to be added.
        """
        self.register_callback("before_import", callback)

    def add_before_load_reference_callback(self, callback):
        """Add a callback to be executed before a reference file is loaded.

        Args:
            callback (function): The callback function to be added.
        """
        self.register_callback("before_load_reference", callback)

    def add_before_import_reference_callback(self, callback):
        """Add a callback to be executed before a reference file is imported.

        Args:
            callback (function): The callback function to be added.
        """
        self.register_callback("before_import_reference", callback)

    def add_maya_exiting_callback(self, callback):
        """Add a callback to be executed when Maya is exiting.

        Args:
            callback (function): The callback function to be added.
        """
        self.register_callback("maya_exiting", callback)

    def add_additionally_fix_function(self, func):
        """Add an additional fix function to be executed.

        Args:
            func (function): The fix function to be added.
        """
        self._additionally_fix_funcs.append(func)

    def collect(self):
        """Collect issues from all loaded vaccines."""
        self.reset()
        for vaccine in self.vaccines:
            vaccine.collect_issues()

    @property
    def have_issues(self):
        """Check if any issues are found.

        Returns:
            bool: True if any issues are found, False otherwise.
        """
        return any([self.malicious_files, self.infected_nodes, self.infected_script_nodes, self.infected_files,
                    self.infected_script_jobs])

    def report(self):
        """Report all issues related to the Maya virus."""
        for name in (
            "malicious_files",
            "infected_nodes",
            "infected_script_jobs",
            "infected_files",
            "infected_reference_files"
        ):
            self.logger.info(self.translator.translate(name, name=getattr(self, name)))

    def reset(self):
        """Reset all issues related to the Maya virus."""
        self._malicious_files = []
        self._infected_nodes = []
        self._infected_script_nodes = []
        self._infected_script_jobs = []
        self._infected_files = []
        self._infected_reference_files = []
        self._registered_callbacks = defaultdict(list)
        self._additionally_fix_funcs = []
